use std::{
    path::{Path, PathBuf},
    sync::{Arc, Mutex, MutexGuard},
};

use serde::Serialize;
use time::{macros::format_description, OffsetDateTime};

use crate::{event::BrokerEvent, run::Run};

#[derive(Debug, Serialize)]
struct StatusData {
    timestamp: String,
    broker_event_counter: usize,
    ci_broker_version: &'static str,
    ci_broker_git_commit: &'static str,
    latest_broker_event: Option<BrokerEvent>,
    latest_ci_run: Option<Run>,
}

impl Default for StatusData {
    fn default() -> Self {
        Self {
            timestamp: "".into(),
            broker_event_counter: 0,
            ci_broker_version: env!("CARGO_PKG_VERSION"),
            ci_broker_git_commit: env!("GIT_HEAD"),
            latest_broker_event: None,
            latest_ci_run: None,
        }
    }
}

impl StatusData {
    fn write(&self, filename: &Path) -> Result<(), StatusError> {
        let tmp = filename.with_extension("update");
        let s = serde_json::to_string_pretty(&self).map_err(StatusError::serialize)?;
        std::fs::write(&tmp, s.as_bytes()).map_err(|e| StatusError::status_write(filename, e))?;
        std::fs::rename(&tmp, filename).map_err(|e| StatusError::status_rename(filename, e))?;
        Ok(())
    }
}

pub struct Status {
    filename: PathBuf,
    status: Arc<Mutex<StatusData>>,
}

impl Status {
    pub fn new(filename: &Path) -> Self {
        Self {
            filename: filename.into(),
            status: Arc::new(Mutex::new(StatusData::default())),
        }
    }

    fn lock(&mut self) -> MutexGuard<StatusData> {
        self.status.lock().expect("lock StatusGuard::status")
    }

    pub fn broker_event(&mut self, event: &BrokerEvent) {
        let mut status = self.lock();
        status.latest_broker_event = Some(event.clone());
        status.broker_event_counter += 1;
    }

    pub fn ci_run(&mut self, run: &Run) {
        let mut status = self.lock();
        status.latest_ci_run = Some(run.clone());
    }

    pub fn write(&mut self) -> Result<(), StatusError> {
        let filename = self.filename.clone();
        let mut status = self.lock();
        status.timestamp = Self::now()?;
        status.write(&filename)?;
        Ok(())
    }

    fn now() -> Result<String, StatusError> {
        let fmt = format_description!("[year]-[month]-[day] [hour]:[minute]:[second]Z");
        OffsetDateTime::now_utc()
            .format(fmt)
            .map_err(StatusError::format_now)
    }
}

impl Clone for Status {
    fn clone(&self) -> Self {
        Self {
            filename: self.filename.clone(),
            status: Arc::clone(&self.status),
        }
    }
}

#[derive(Debug, thiserror::Error)]
pub enum StatusError {
    #[error("failed to format current time stamp")]
    FormatNow(#[source] time::error::Format),

    #[error("failed to serialize status as JSON")]
    Serizalize(#[source] serde_json::Error),

    #[error("failed to write status to file {0}")]
    StatusWrite(PathBuf, #[source] std::io::Error),

    #[error("failed to rename status to file {0}")]
    StatusRename(PathBuf, #[source] std::io::Error),
}

impl StatusError {
    fn format_now(err: time::error::Format) -> Self {
        Self::FormatNow(err)
    }

    fn serialize(err: serde_json::Error) -> Self {
        Self::Serizalize(err)
    }

    fn status_write(filename: &Path, err: std::io::Error) -> Self {
        Self::StatusWrite(filename.into(), err)
    }

    fn status_rename(filename: &Path, err: std::io::Error) -> Self {
        Self::StatusRename(filename.into(), err)
    }
}
